/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	Functions implementing Event abstractions in the Linux OS.
 *
 * Environment
 *	Linux Kernel Mode
 *
 * $Revision: 1.1 $
 */

#include <complib/cl_event.h>
#include <complib/cl_memory.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>

/*
 * Internal functions needed for events
 */
typedef struct _osd_timeout {
	struct timer_list timer;
	boolean_t timedout;
	wait_queue_head_t *queue;

} osd_timeout_t;

void
__osd_timeout_handler(unsigned long arg)
{
	osd_timeout_t *t = (osd_timeout_t *) arg;

	t->timedout = TRUE;
	wake_up(t->queue);
}

void
__osd_set_timeout(osd_timeout_t * t, int milliseconds, wait_queue_head_t * q)
{
	t->timedout = FALSE;
	t->queue = q;
	init_timer(&t->timer);

	/* HZ is the number of ticks per second for Intel architectures. */
	t->timer.expires = jiffies + (milliseconds * HZ) / 1000;
	t->timer.data = (unsigned long) t;
	t->timer.function = __osd_timeout_handler;
	add_timer(&t->timer);
}

/* Deletes a timer and returns whether that timer had timed out. */
boolean_t
__osd_clear_timeout(osd_timeout_t * t)
{
	del_timer(&t->timer);
	return (t->timedout);
}

/* Declare the we intended to create an event */
void
cl_event_construct(IN cl_event_t * const p_event)
{
	CL_ASSERT(p_event);

	cl_memclr(p_event, sizeof (cl_event_t));
	cl_spinlock_construct(&p_event->spinlock);
	p_event->state = CL_UNINITIALIZED;
}

/* Initialize an event */
cl_status_t
cl_event_init(IN cl_event_t * const p_event, IN const boolean_t manual_reset)
{
	cl_status_t status;

	CL_ASSERT(p_event);

	cl_event_construct(p_event);

	init_waitqueue_head(&p_event->wait_queue);
	status = cl_spinlock_init(&p_event->spinlock);
	if (status != CL_SUCCESS)
		return (status);

	p_event->signaled = FALSE;
	p_event->manual_reset = manual_reset;
	p_event->state = CL_INITIALIZED;
	return (CL_SUCCESS);
}

/* Destroy an event */
void
cl_event_destroy(IN cl_event_t * const p_event)
{
	CL_ASSERT(p_event);
	CL_ASSERT(cl_is_state_valid(p_event->state));

	cl_spinlock_destroy(&p_event->spinlock);
	p_event->state = CL_UNINITIALIZED;
}

/* Trigger or Signal an event */
cl_status_t
cl_event_signal(IN cl_event_t * const p_event)
{
	CL_ASSERT(p_event);
	/* Make sure that the event was started */
	CL_ASSERT(p_event->state == CL_INITIALIZED);

	cl_spinlock_acquire(&p_event->spinlock);
	p_event->signaled = TRUE;
	if (waitqueue_active(&p_event->wait_queue)) {
		if (p_event->manual_reset) {
			/* Wake up all waiters. */
			wake_up_all(&p_event->wait_queue);
		} else {
			/* Wake up one waiter. */
			wake_up(&p_event->wait_queue);
			/* Reset the event. */
			p_event->signaled = FALSE;
		}
	}
	cl_spinlock_release(&p_event->spinlock);

	return (CL_SUCCESS);
}

/* Clear an event, as though it was never triggered */
cl_status_t
cl_event_reset(IN cl_event_t * const p_event)
{
	CL_ASSERT(p_event);
	/* Make sure that the event was started */
	CL_ASSERT(p_event->state == CL_INITIALIZED);

	cl_spinlock_acquire(&p_event->spinlock);
	p_event->signaled = FALSE;
	cl_spinlock_release(&p_event->spinlock);

	return (CL_SUCCESS);
}

/* Wait for an event */
cl_status_t
cl_event_wait_on(IN cl_event_t * const p_event,
		 IN const uint32_t wait_us, IN const boolean_t interruptible)
{
	cl_status_t ret;
	osd_timeout_t timeout;
	int32_t milliseconds;
	DECLARE_WAITQUEUE(wait, current);

	CL_ASSERT(p_event);
	/* Make sure that the event was Started */
	CL_ASSERT(p_event->state == CL_INITIALIZED);

	cl_spinlock_acquire(&p_event->spinlock);

	/* Return immediately if the event is signalled. */
	if (p_event->signaled) {
		if (!p_event->manual_reset)
			p_event->signaled = FALSE;

		cl_spinlock_release(&p_event->spinlock);
		return (CL_SUCCESS);
	}

	/* If just testing the state, return CL_TIMEOUT. */
	if (wait_us == 0) {
		cl_spinlock_release(&p_event->spinlock);
		return (CL_TIMEOUT);
	}

	/*
	 * We will be performing a wait, and should not be in an
	 * interrupt handler.
	 * Since we use spin_lock_bh() in_interrupt returns true! hence
	 * commented below safety check.
	 * CL_ASSERT( !in_interrupt() );
	 */

	add_wait_queue_exclusive(&p_event->wait_queue, &wait);

	/* Set the interruptible state as requested. */
	if (interruptible)
		set_current_state(TASK_INTERRUPTIBLE);
	else
		set_current_state(TASK_UNINTERRUPTIBLE);

	
	/* If a timeout was specified, set a timer to wake us up. */
	if (wait_us != EVENT_NO_TIMEOUT) {
		/* Linux timers provide millisecond accuracy. */
		milliseconds = wait_us / 1000;
		/* Minimum wait is 10 millisecond. */
		if (milliseconds < 10)
			milliseconds = 10;

		__osd_set_timeout(&timeout, milliseconds, &p_event->wait_queue);
	}

	cl_spinlock_release(&p_event->spinlock);
	/* Let other threads run. */
	schedule();

	cl_spinlock_acquire(&p_event->spinlock);
	remove_wait_queue(&p_event->wait_queue, &wait);
	cl_spinlock_release(&p_event->spinlock);

	/*
	 * If we agreed to be alerted and there is a signal pending, we
	 * got woken up because of a signal. This is different from being
	 * woken up because of the event we are waiting on or being timed out.
	 */

	if (interruptible && signal_pending(current)) {
		/*
		 *  If we woke up due to a pending signal, then
		 *  clear the timer that was running
		 */
		if (wait_us != EVENT_NO_TIMEOUT)
			__osd_clear_timeout(&timeout);

		ret = CL_NOT_DONE;
	} else if (wait_us != EVENT_NO_TIMEOUT && __osd_clear_timeout(&timeout)) {
		ret = CL_TIMEOUT;
	} else
		ret = CL_SUCCESS;

	if (!p_event->manual_reset) {
		if (cl_event_reset(p_event) == CL_ERROR)
			ret = CL_ERROR;
	}

	return (ret);
}
